import pygame
from pygame.locals import *

import pathfind
import random
import math

import numpy as np
from collections import deque

from random import choice

from avatar import *
  
level = pathfind.level1

# List of hiding locations.
hiding = [(0,2), (4,1), (2,0)]

# Small tweak to the game loop when running in IDLE to avoid
# killing the IDLE window.
running = True

pygame.init()

screen = pygame.display.set_mode((640,480), 0, 32)

# To easily construct bounding boxes, images are now going to be rectangles.
boy_image = pygame.Surface((25,25))
boy_image.fill((135,145,191))
                    
girl_image = pygame.Surface((25,25))
girl_image.fill((255,192,203))

# Scratch location for data.
enemy_bb = {
  'health' : 100,
  'ammo' : 10,
  'time_passed' : 0,
  'delay': 0,
  'state': 'None',
  'panic': None,
  'home': (550.0, 350.0)
}

player_bb = {
  'health' : 100,
  'ammo' : 20,
  'home': (50.0, 50.0),
  'state' : 'None',
  'hiding' : False
}

def mandist(t, u):
  tx, ty = t
  ux, uy = u
  return abs(tx - ux) + abs(ty - uy)

# TODO Begin State Logic
class State(object):

  def __init__(self, ai):
    self.ai = ai

  def getEntryAction(self):
    pass

  def getExitAction(self):
    pass

  def getAction(self):
    pass

  def changeState(self, nextState):
    self.getExitAction()
    self.ai.state = nextState
    nextState.getEntryAction()
# TODO End State Logic

boy = Avatar("Boy", boy_image, (50.0,50.0), 100.0, player_bb)
girl = Avatar("Girl", girl_image, (550.0, 350.0), 150.0, enemy_bb)

# Add boy to the girl's blackboard.
girl.blackboard['player'] = boy

# TODO: Initial AI State
# girl.state = WaitState(girl)
girl.state = State(girl)

clock = pygame.time.Clock()

health_image = pygame.Surface((10,10))
health_image.fill((0,255,0))

ammo_image = pygame.Surface((10,10))
ammo_image.fill((255,0,0))

# Load an array of tiles to translate the level to a block.
tiles = {
  ".": (50,50,50),
  "#": (128,0,0)  
}

# Knowledge is stored in the blackboard.
def grabItem(avatar):

  if avatarPositionToTile(avatar) == items['health']:

    items['health'] = None
    avatar.blackboard['health'] = avatar.blackboard['health'] + 10
    
    if avatar.blackboard['health'] > 100:
      avatar.blackboard['health'] = 100

  if avatarPositionToTile(avatar) == items['ammo']:

    items['ammo'] = None
    avatar.blackboard['ammo'] = avatar.blackboard['ammo'] + 10
    
    if avatar.blackboard['ammo'] > 100:
      avatar.blackboard['ammo'] = 100
  
# The height and width of each tile.
TILE_WIDTH = 100
TILE_HEIGHT = 100

# Low-level seek movement; ai should be an avatar object.
def seek(ai, to, time_passed):
  target = to - ai.position
  ai.update(target, time_passed)

# Player has noclip. This function takes a tile coordinate
# and converts it to pixels so that the pixels are in
# the center of the tile.
def tilePositionToPixel(tile):

  return (np.array((TILE_WIDTH / 2, TILE_HEIGHT / 2)) +
          np.array((TILE_WIDTH, TILE_HEIGHT)) * tile)

def avatarPositionToTile(avatar):
  x, y = avatar.position
  return (int(x) / TILE_WIDTH, int(y) / TILE_HEIGHT)

def findNext(level, ai, target):

    # Figure out which tile you and your enemy are by
    # unpacking the player.    
    ai_x, ai_y = ai.position
    target_x, target_y = target

    player_tile = (int(target_x) / TILE_WIDTH, int(target_y) / TILE_HEIGHT)
    enemy_tile = (int(ai_x)/ TILE_WIDTH, int(ai_y) / TILE_HEIGHT)

    path = pathfind.bfs(level, enemy_tile, player_tile)

    if path == None:
      return None
    if len(path) == 1:
      return tilePositionToPixel(path[0])
    else:
      return tilePositionToPixel(path[1])      
  
def drawBoard(level, tiles, screen):

  global hiding
  
  width = len(level) * TILE_WIDTH
  height = len(level[0]) * TILE_HEIGHT
 
  for x in range(len(level[0])):
    for y in range(len(level)):

      where = (x * TILE_WIDTH, y * TILE_HEIGHT, TILE_WIDTH, TILE_HEIGHT)
      color = tiles[level[y][x]]

      if level[y][x] == '#':
        pygame.draw.rect(screen, color, Rect(where))
      else:
        pygame.draw.rect(screen, color, Rect(where), 1)

  for h in hiding:    
    x, y = h

    where = (x * TILE_WIDTH, y * TILE_HEIGHT, TILE_WIDTH, TILE_HEIGHT)
    color = (255, 255, 204)

    pygame.draw.rect(screen, color, Rect(where))
    
            
def statusline(screen):

  # Fancy font handling.
  status = 'P'

  font = pygame.font.SysFont("Consolas", 20)
  text = font.render(status, True, (255,255,255), (0,0,0))

  textRect = text.get_rect()
  textRect.topleft = (10,450)

  screen.blit(text, textRect)

  status = 'E | {}'.format(girl.blackboard['state'])

  text = font.render(status, True, (255,255,255), (0,0,0))
  textRect = text.get_rect()
  textRect.topright = (630, 450)

  screen.blit(text, textRect)

# Bootstrap AI
girl.state.getEntryAction()
      
while running:

  time_passed = clock.tick(30)
  time_passed_seconds = time_passed / 1000.0

  # Slow down the AI  
  girl.blackboard['time_passed'] = time_passed_seconds  
  girl.blackboard['delay'] = girl.blackboard['delay'] + time_passed_seconds

  # Cutscenes
  boy.blackboard['time_passed'] = time_passed_seconds

  for event in pygame.event.get():
    
    if (event.type == QUIT or
        (event.type == KEYDOWN and event.key == K_ESCAPE)):

      running = False
  
  pressed_keys = pygame.key.get_pressed()

  target = np.array([0.0,0.0])

  # TODO: Hide functionality.
    
  if pressed_keys[K_LEFT]:
    target += np.array([-1.0,0.0])
  if pressed_keys[K_RIGHT]:
    target += np.array([1.0,0.0])
  if pressed_keys[K_UP]:
    target += np.array([0.0,-1.0])
  if pressed_keys[K_DOWN]:
    target += np.array([0.0,1.0])

  boy.update(target, time_passed_seconds)

  screen.fill((0,0,0))
    
  drawBoard(level, tiles, screen)
  statusline(screen)
  
  # Needs to be repositioned.
  screen.blit(girl.surface, girl.position - girl.surface.get_rect().center)
  screen.blit(boy.surface, boy.position - girl.surface.get_rect().center)  
  
  # FSM
  girl.state.getAction()
    
  pygame.display.update()

# Wrap up.
pygame.quit()
